import argparse

import torch
import numpy as np
from src.utils import prepare_data, prepare_data_typical
from src.model import LinkPred
from sklearn.metrics import roc_auc_score, average_precision_score
from tqdm import tqdm

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print('Device:', device)


def save_file(scores: list, fold_idx: int, name: str):
    file_name = f'{name}_result_fold_{fold_idx}.txt'
    with open(file_name, 'w') as f:
        for n1, n2, s in scores:
            f.write(f'{n1}\t{n2}\t{s}\n')


def set_random_seed(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--data_name', type=str, default='chemical', help='what data is used for experiment.')
    # data_name : the name of dataset, which are atc, chemical, Cdataset, Gdataset, and lrssl
    parser.add_argument('--num_hops', type=int, default=2, help='the number of hops for k-hop neighborhood sampling')
    parser.add_argument('--init_attribute', type=str, default='ones', help='the initial feature used for the node')
    parser.add_argument('--init_representation', type=str, default=None, help='generating represents of nodes using node2vec or etc. default=None')
    parser.add_argument('--max_nodes_per_hop', type=int, default=None, help='the max number of neighbors for each hop')
    parser.add_argument('--embedding_dim', type=int, default=32, help='the size of embedding dimension for generating initial representation')
    parser.add_argument('--drnl', type=bool, default=True, help='the use of DRNL scheme')
    parser.add_argument('--seed', type=int, default=1, help='the number for random seed')
    parser.add_argument('--lr', type=float, default=0.00005, help='learning rate')
    parser.add_argument('--weight_decay', type=float, default=0, help='the weight decay for adam optimizer')
    parser.add_argument('--walk_len', type=int, default=7, help='the length of walking for random walk')
    parser.add_argument('--heads', type=int, default=2, help='the number of heads used for multi-head attention')
    parser.add_argument('--hidden_channels', type=int, default=32, help='the size of hidden channels for deep learning model')
    parser.add_argument('--batch_size', type=int, default=32, help='the batch size')
    parser.add_argument('--epoch_num', type=int, default=50, help='the number of epochs')
    parser.add_argument('--gpu_multi', type=bool, default=False, help='the use multiple gpus')
    parser.add_argument('--fold_side', type=str, default='drug', help='the side for cross validation. drug or disease')
    parser.add_argument('--garnet', type=bool, default=True, help='the use of GARNET model.')
    parser.add_argument('--disease_rank', type=int, default=90, help='the rank for reduced-rank SVD model used for disease network')
    parser.add_argument('--drug_rank', type=int, default=90, help='the rank for reduced-rank SVD model used for drug network')
    parser.add_argument('--num_k', type=int, default=1, help='the number of neighbors for kNN used in GARNET')

    args = parser.parse_args()

    if args.data_name == 'atc' and args.fold_side == 'drug':
        args.disease_rank, args.drug_rank, args.num_k = 15, 30, 6
    elif args.data_name == 'atc' and args.fold_side == 'disease':
        args.disease_rank, args.drug_rank, args.num_k = 23, 30, 6
    elif args.data_name == 'chemical' and args.fold_side == 'drug':
        args.disease_rank, args.drug_rank, args.num_k = 3, 16, 6
    elif args.data_name == 'chemical' and args.fold_side == 'disease':
        args.disease_rank, args.drug_rank, args.num_k = 20, 25, 6

    # 10-fold Cross Validation
    for i in range(10):
        args.fold_idx = i
        print("-" * 50 + 'Dataset and Features' + "-" * 60)
        print("{:<10}|{:<10}|{:<10}|{:<10}|{:<10}|{:<15}" \
              .format('Dataset', 'Dimension', 'Attribute', 'hops', 'DRNL', 'Representation'))
        print("-" * 130)
        print("{:<10}|{:<10}|{:<10}|{:<10}|{:<10}|{:<15}" \
              .format(args.data_name, args.embedding_dim, str(args.init_attribute), str(args.num_hops), str(args.drnl), str(args.init_representation)))
        print("-" * 130)

        print("-" * 42 + 'Model and Training' + "-" * 45)
        print("{:<13}|{:<13}|{:<13}|{:<8}|{:<13}|{:<8}|{:<15}" \
              .format('Learning Rate', 'Weight Decay', 'Batch Size', 'Epoch', 'Walk Length', 'Heads', 'Hidden Channels'))
        print("-" * 105)

        print("{:<13}|{:<13}|{:<13}|{:<8}|{:<13}|{:<8}|{:<15}" \
              .format(args.lr, args.weight_decay, str(args.batch_size), \
                      args.epoch_num, args.walk_len, args.heads, args.hidden_channels))
        print("-" * 105)

        print('<<Begin generating training data>>')

        if args.data_name in ['Cdataset', 'Gdataset', 'lrssl']:
            train_loader, test_loader, feature_results = prepare_data_typical(args)
        else:
            train_loader, test_loader, feature_results = prepare_data(args)

        print('<<Complete generating training data>>')

        walk_len = args.walk_len
        heads = args.heads
        hidden_channels = args.hidden_channels
        lr = args.lr
        weight_decay = args.weight_decay

        torch.cuda.empty_cache()

        num_features = next(iter(train_loader)).x.size(1)

        z_max = 0
        if args.drnl == True:
            for data in train_loader:
                z_max = max(z_max, torch.max(data.z).numpy())
            # for data in val_loader:
            #     z_max = max(z_max, torch.max(data.z).numpy())
            for data in test_loader:
                z_max = max(z_max, torch.max(data.z).numpy())
            z_max = z_max + 1

            # if use drnl, we use a Embedding with dimension = hidden_channels
            num_features = hidden_channels + num_features

        torch.cuda.empty_cache()
        print("Dimension of features after concatenation:", num_features)
        set_random_seed(args.seed)

        model = LinkPred(in_channels=num_features, hidden_channels=hidden_channels, \
                         heads=heads, walk_len=walk_len, drnl=args.drnl, z_max=z_max).to(device)

        optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
        criterion = torch.nn.BCEWithLogitsLoss()

        def train(loader, epoch):
            model.train()
            loss_epoch = 0
            for data in tqdm(loader, desc="train"):  # Iterate in batches over the training dataset.
                data = data.to(device)
                label = data.label
                out = model(data.x, data.edge_index, data.edge_mask, data.batch, data.z)
                torch.cuda.empty_cache()
                loss = criterion(out.view(-1), label)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                loss_epoch = loss_epoch + loss.item()
            return loss_epoch / len(loader)


        def test(loader, data_type='test'):
            model.eval()
            original_score = []
            scores = torch.tensor([])
            labels = torch.tensor([])
            loss_total = 0
            with torch.no_grad():
                # for data in tqdm(loader,position=0,leave=True):  # Iterate in batches over the training/test dataset.
                for data in tqdm(loader, desc='test:' + data_type):  # Iterate in batches over the training/test dataset.
                    data = data.to(device)
                    out = model(data.x, data.edge_index, data.edge_mask, data.batch, data.z)
                    loss = criterion(out.view(-1), data.label)
                    out = out.cpu().clone().detach()
                    scores = torch.cat((scores, out), dim=0)
                    labels = torch.cat((labels, data.label.view(-1, 1).cpu().clone().detach()), dim=0)
                    for i in range(data.num_graphs):
                        original_edge = data.get_example(i).original_edge.tolist()
                        original_score.append((original_edge[0], original_edge[1], out.tolist()[i][0]))
                scores = scores.cpu().clone().detach().numpy()
                labels = labels.cpu().clone().detach().numpy()
                loss_total = loss_total + loss.item()
                return roc_auc_score(labels, scores), average_precision_score(labels, scores), loss_total, original_score


        best_auc = 0
        best_ap = 0
        best_score_value = []
        best_epoch = 0

        for epoch in range(0, args.epoch_num):
            loss_epoch = train(train_loader, epoch)
            # val_auc, val_ap, val_loss = test(val_loader, data_type='val')
            test_auc, test_ap, _, score_value = test(test_loader, data_type='test')
            if test_auc > best_auc:
                best_auc = test_auc
                best_score_value = score_value
                best_epoch = epoch
                best_ap = test_ap

            print(f'Fold_idx: {args.fold_idx}, Epoch: {epoch:03d}, Loss : {loss_epoch:.4f}, Test AUC: {test_auc:.4f}, Picked AUC:{best_auc:.4f}')

        print(f'Best test AUC: {best_auc:.4f}, Best test AP: {best_ap:.4f}')
        save_file(best_score_value, args.fold_idx, args.data_name)


