from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
import torch
from sklearn.model_selection import KFold
import numpy as np
from torch_geometric.utils import dense_to_sparse
import torch.nn.functional as F
import sys
import os.path
import networkx as nx
from src.drnl import drnl_node_labeling
from src.reduced_SVD import rrSVD
import scipy.io as sio
import pandas as pd
from scipy.sparse import csr_matrix


cur_dir = os.path.dirname(os.path.realpath(__file__))
par_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append('%s/software/' % par_dir)

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


def floor(x):
    return torch.div(x, 1, rounding_mode='trunc')


def set_random_seed(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def k_hop_subgraph(node_idx, num_hops, edge_index, max_nodes_per_hop=None, num_nodes=None):

    if num_nodes == None:
        num_nodes = torch.max(edge_index)+1
    row, col = edge_index
    node_mask = row.new_empty(num_nodes, dtype=torch.bool)
    edge_mask = row.new_empty(row.size(0), dtype=torch.bool)

    node_idx = node_idx.to(row.device)

    subsets = [node_idx]

    if max_nodes_per_hop == None:
        for _ in range(num_hops):
            node_mask.fill_(False)
            node_mask[subsets[-1]] = True
            torch.index_select(node_mask, 0, row, out = edge_mask)
            subsets.append(col[edge_mask])
    else:
        not_visited = row.new_empty(num_nodes, dtype=torch.bool)
        not_visited.fill_(True)
        for _ in range(num_hops):
            node_mask.fill_(False)# the source node mask in this hop
            node_mask[subsets[-1]] = True #mark the sources
            not_visited[subsets[-1]] = False # mark visited nodes
            torch.index_select(node_mask, 0, row, out = edge_mask) # indices of all neighbors
            neighbors = col[edge_mask].unique() #remove repeats
            neighbor_mask = row.new_empty(num_nodes, dtype=torch.bool) # mask of all neighbor nodes
            edge_mask_hop = row.new_empty(row.size(0), dtype=torch.bool) # selected neighbor mask in this hop
            neighbor_mask.fill_(False)
            neighbor_mask[neighbors] = True
            neighbor_mask = torch.logical_and(neighbor_mask, not_visited) # all neighbors that are not visited
            ind = torch.where(neighbor_mask==True) #indicies of all the unvisited neighbors
            if ind[0].size(0) > max_nodes_per_hop:
                perm = torch.randperm(ind[0].size(0))
                ind = ind[0][perm]
                neighbor_mask[ind[max_nodes_per_hop:]] = False # randomly select max_nodes_per_hop nodes
                torch.index_select(neighbor_mask, 0, col, out = edge_mask_hop)# find the indicies of selected nodes
                edge_mask = torch.logical_and(edge_mask,edge_mask_hop) # change edge_mask
            subsets.append(col[edge_mask])

    subset, inv = torch.cat(subsets).unique(return_inverse=True)
    inv = inv[:node_idx.numel()]

    node_mask.fill_(False)
    node_mask[subset] = True
    edge_mask = node_mask[row] & node_mask[col]

    edge_index = edge_index[:, edge_mask]

    node_idx = row.new_full((num_nodes, ), -1)
    node_idx[subset] = torch.arange(subset.size(0), device=row.device)
    edge_index = node_idx[edge_index]

    return subset, edge_index, inv, edge_mask


def plus_edge(data_observed, label, p_edge, args):
    nodes, edge_index_m, mapping, _ = k_hop_subgraph(node_idx= p_edge, num_hops=args.num_hops,\
 edge_index = data_observed.edge_index, max_nodes_per_hop=args.max_nodes_per_hop ,num_nodes=data_observed.num_nodes)
    x_sub = data_observed.x[nodes,:]
    edge_index_p = edge_index_m
    edge_index_p = torch.cat((edge_index_p, mapping.view(-1,1)),dim=1)
    edge_index_p = torch.cat((edge_index_p, mapping[[1,0]].view(-1,1)),dim=1)

    #edge_mask marks the edge under perturbation, i.e., the candidate edge for LP
    edge_mask = torch.ones(edge_index_p.size(1),dtype=torch.bool)
    edge_mask[-1] = False
    edge_mask[-2] = False

    if args.drnl == True:
        num_nodes = torch.max(edge_index_p)+1
        z = drnl_node_labeling(edge_index_m, mapping[0],mapping[1],num_nodes)
        data = Data(edge_index = edge_index_p, x = x_sub, z = z)
    else:
        data = Data(edge_index = edge_index_p, x = x_sub, z = 0)
    data.edge_mask = edge_mask
    data.original_edge = p_edge

    #label = 1 if the candidate link (p_edge) is positive and label=0 otherwise
    data.label = float(label)

    return data


def minus_edge(data_observed, label, p_edge, args):
    nodes, edge_index_p, mapping,_ = k_hop_subgraph(node_idx= p_edge, num_hops=args.num_hops,\
 edge_index = data_observed.edge_index,max_nodes_per_hop=args.max_nodes_per_hop, num_nodes = data_observed.num_nodes)
    x_sub = data_observed.x[nodes,:]

    #edge_mask marks the edge under perturbation, i.e., the candidate edge for LP
    edge_mask = torch.ones(edge_index_p.size(1), dtype = torch.bool)
    ind = torch.where((edge_index_p == mapping.view(-1,1)).all(dim=0))
    edge_mask[ind[0]] = False
    ind = torch.where((edge_index_p == mapping[[1,0]].view(-1,1)).all(dim=0))
    edge_mask[ind[0]] = False
    if args.drnl == True:
        num_nodes = torch.max(edge_index_p)+1
        z = drnl_node_labeling(edge_index_p[:,edge_mask], mapping[0],mapping[1],num_nodes)
        data = Data(edge_index = edge_index_p, x= x_sub,z = z)
    else:
        data = Data(edge_index = edge_index_p, x= x_sub,z = 0)
    data.edge_mask = edge_mask
    data.original_edge = p_edge

    #label = 1 if the candidate link (p_edge) is positive and label=0 otherwise
    data.label = float(label)
    return data


def set_init_attribute_representation(data,args):
    # Construct data_observed and compute its node attributes & representation
    edge_index = torch.cat((data.train_pos,data.train_pos[[1,0],:]),dim=1)
    data_observed = Data(edge_index=edge_index)
    data_observed.num_nodes = data.num_nodes
    edge_index_observed = data_observed.edge_index

    # generate the initial node attribute if there isn't any
    if data.x == None:
        if args.init_attribute =='one_hot':
            x = F.one_hot(torch.arange(data.num_nodes), num_classes=data.num_nodes)
            x = x.float()
        if args.init_attribute =='ones':
            x = torch.ones(data.num_nodes,args.embedding_dim)
            x = x.float()
        if args.init_attribute =='zeros':
            x = torch.zeros(data.num_nodes,args.embedding_dim)
            x = x.float()
    else:
        x = data.x

    data_observed.x = x
    feature_results = None

    return data_observed,feature_results


def prepare_data(args):
    file_path = f'data/{args.data_name}'
    num_nodes = 921  # Here, the number of node, should be changed
    if args.data_name == 'chemical':
        num_nodes = 944
    num_diseases = 285
    num_drugs = int(num_nodes) - num_diseases

    # for GARNET
    disease_graph = nx.Graph()
    drug_graph = nx.Graph()
    for i in range(num_diseases):
        disease_graph.add_node(i)
    for i in range(num_drugs):
        drug_graph.add_node(i)

    # di-network
    disease_edges = []
    with open(f'{file_path}/edge_list_idx_sorted_di.txt') as di_f:
        for line in di_f:
            line = line.strip().split('\t')
            disease_graph.add_edge(int(line[0]), int(line[1]), weight=float(line[2]))
    #         disease_edges.append((int(line[0]), int(line[1]), float(line[2])))
    #
    # disease_edges.sort(key=lambda x: x[2], reverse=True)
    # num_disease_edges = int(args.density * 0.01 * num_diseases * (num_diseases - 1) / 2)
    # disease_edges = disease_edges[:num_disease_edges]
    # for n1, n2, w in disease_edges:
    #     disease_graph.add_edge(n1, n2, weight=w)

    # dr-network
    drug_edges = []
    with open(f'{file_path}/edge_list_idx_sorted_dr.txt') as dr_f:
        for line in dr_f:
            line = line.strip().split('\t')
            drug_graph.add_edge(int(line[0]) - num_diseases, int(line[1]) - num_diseases, weight=float(line[2]))
    #         drug_edges.append((int(line[0]) - num_diseases, int(line[1]) - num_diseases, float(line[2])))
    # drug_edges.sort(key=lambda x: x[2], reverse=True)
    # num_drug_edges = int(args.density * 0.01 * num_drugs * (num_drugs - 1) / 2)
    # drug_edges = drug_edges[:num_drug_edges]
    # for n1, n2, w in drug_edges:
    #     drug_graph.add_edge(n1, n2, weight=w)

    # for GARNET
    # convert networkx graph to scipy sparse matrix
    disease_adj = nx.adjacency_matrix(disease_graph)
    drug_adj = nx.adjacency_matrix(drug_graph)

    # adopt GARNET
    rr_disease_adj = rrSVD(disease_adj, None, r=args.disease_rank, k=args.num_k, use_feature=False)
    rr_drug_adj = rrSVD(drug_adj, None, r=args.drug_rank, k=args.num_k, use_feature=False)

    rr_disease_graph = nx.from_scipy_sparse_matrix(rr_disease_adj)
    rr_drug_graph = nx.from_scipy_sparse_matrix(rr_drug_adj)

    G = nx.Graph()
    for i in range(num_nodes):
        G.add_node(i)

    for n1, n2 in list(rr_disease_graph.edges):
        G.add_edge(n1, n2)

    for n1, n2 in list(rr_drug_graph.edges):
        G.add_edge(n1 + 285, n2 + 285)

    # di-dr edge
    with open(f'{file_path}/edge_list_idx_sorted_didr.txt') as didr_f:
        num_nodes = didr_f.readline()
        for i in range(int(num_nodes)):
            G.add_node(i)
        for line in didr_f:
            line = line.strip().split('\t')
            G.add_edge(int(line[0]), int(line[1]))

    adj = nx.adjacency_matrix(G, sorted(G.nodes()))
    A = adj.toarray()
    edge_index, _ = dense_to_sparse(torch.tensor(A))
    data = Data(edge_index=edge_index)
    data = split_edges(data, args)

    set_random_seed(args.seed)
    data_observed, feature_results = set_init_attribute_representation(data, args)

    # Construct train, val and test data loader.
    set_random_seed(args.seed)
    train_graphs = []
    test_graphs = []
    for i in range(data.train_pos.size(1)):
        train_graphs.append(minus_edge(data_observed, 1, data.train_pos[:, i], args))

    for i in range(data.train_neg.size(1)):
        train_graphs.append(plus_edge(data_observed, 0, data.train_neg[:, i], args))

    for i in range(data.test_pos.size(1)):
        test_graphs.append(plus_edge(data_observed, 1, data.test_pos[:, i], args))

    for i in range(data.test_neg.size(1)):
        test_graphs.append(plus_edge(data_observed, 0, data.test_neg[:, i], args))

    print('Train_link:', str(len(train_graphs)), ' Test_link:', str(len(test_graphs)))

    if args.gpu_multi:
        return train_graphs, test_graphs, feature_results

    train_loader = DataLoader(train_graphs, batch_size=args.batch_size, shuffle=True)
    test_loader = DataLoader(test_graphs, batch_size=args.batch_size, shuffle=False)

    return train_loader, test_loader, feature_results


# 0 ~ 284 : disease // 285 ~ 920 : drug
def split_edges(data, args):
    num_disease = 285
    end_disease_idx = num_disease - 1
    num_nodes = data.num_nodes
    folds_file_path = os.path.join(par_dir, 'data', args.data_name, f'{args.fold_side}_10-fold_idx.txt')
    folds = []
    with open(folds_file_path) as f:
        for line in f:
            folds.append(list(map(int, line.strip().split('\t'))))

    fold = folds[args.fold_idx]
    edge_index = data.edge_index

    row, col = data.edge_index
    mask = row < col
    row, col = row[mask], col[mask]
    edge_index_train = []
    test_pos_set = set()

    if args.fold_side == 'drug':
        for i in range(row.size(0)):
            n1, n2 = row[i], col[i]
            if n2 in fold and n1 <= end_disease_idx:
                test_pos_set.add((int(n1), int(n2)))
            else:
                edge_index_train.append(torch.stack([n1, n2], dim=0))
        data.train_pos = torch.stack(edge_index_train, dim=1)

        edge_index_test_pos = []
        edge_index_test_neg = []
        for dr in fold:
            for di in range(num_disease):
                if (di, dr) in test_pos_set:
                    edge_index_test_pos.append(np.array([di, dr]))
                else:
                    edge_index_test_neg.append(np.array([di, dr]))
        edge_index_test_pos = np.array(edge_index_test_pos)
        edge_index_test_neg = np.array(edge_index_test_neg)
        data.test_pos = torch.tensor(edge_index_test_pos).T.type(torch.int64)
        data.test_neg = torch.tensor(edge_index_test_neg).T.type(torch.int64)

        set_random_seed(args.seed)

        neg_adj_mask = torch.ones(data.num_nodes, data.num_nodes, dtype=torch.uint8)
        neg_adj_mask = neg_adj_mask.triu(diagonal=1).to(torch.bool)
        neg_adj_mask[row, col] = 0

        neg_row, neg_col = neg_adj_mask.nonzero(as_tuple=False).t()
        perm = torch.randperm(neg_row.size(0))[:row.size(0)]
        neg_row, neg_col = neg_row[perm], neg_col[perm]
        row, col = neg_row[:data.train_pos.size(1)], neg_col[:data.train_pos.size(1)]
        data.train_neg = torch.stack([row, col], dim=0)
    else:
        for i in range(row.size(0)):
            n1, n2 = row[i], col[i]
            if n1 in fold and n2 >= end_disease_idx:
                test_pos_set.add((int(n1), int(n2)))
            else:
                edge_index_train.append(torch.stack([n1, n2], dim=0))
        data.train_pos = torch.stack(edge_index_train, dim=1)

        edge_index_test_pos = []
        edge_index_test_neg = []
        for di in fold:
            for dr in range(num_disease+1, num_nodes):
                if (di, dr) in test_pos_set:
                    edge_index_test_pos.append(np.array([di, dr]))
                else:
                    edge_index_test_neg.append(np.array([di, dr]))
        edge_index_test_pos = np.array(edge_index_test_pos)
        edge_index_test_neg = np.array(edge_index_test_neg)
        data.test_pos = torch.tensor(edge_index_test_pos).T.type(torch.int64)
        data.test_neg = torch.tensor(edge_index_test_neg).T.type(torch.int64)

        set_random_seed(args.seed)

        neg_adj_mask = torch.ones(data.num_nodes, data.num_nodes, dtype=torch.uint8)
        neg_adj_mask = neg_adj_mask.triu(diagonal=1).to(torch.bool)
        neg_adj_mask[row, col] = 0

        neg_row, neg_col = neg_adj_mask.nonzero(as_tuple=False).t()
        perm = torch.randperm(neg_row.size(0))[:row.size(0)]
        neg_row, neg_col = neg_row[perm], neg_col[perm]
        row, col = neg_row[:data.train_pos.size(1)], neg_col[:data.train_pos.size(1)]
        data.train_neg = torch.stack([row, col], dim=0)

    return data


# disease nodes + drug nodes
# node index: 0 ~ num_diseases -1 (diseases) / num_diseases ~ num_diseases + num_drugs - 1
def prepare_data_typical(args):
    association_matrix = None
    data_name = args.data_name
    if data_name in ['Gdataset', 'Cdataset']:
        file_path = f'data/Typical/{data_name}/{data_name}.mat'
        data = sio.loadmat(file_path)
        association_matrix = data['didr'].T
        disease_sim_features = data['disease']
        drug_sim_features = data['drug']
    elif data_name in ['lrssl']:
        file_path = f'data/Typical/{data_name}/'
        data = pd.read_csv(os.path.join(file_path, 'drug_dis.txt'), index_col=0, delimiter='\t')
        association_matrix = data.values
        disease_sim_features = pd.read_csv(os.path.join(file_path, 'dis_sim.txt'), index_col=0, delimiter='\t').values
        drug_sim_features = pd.read_csv(os.path.join(file_path, 'drug_sim.txt'), index_col=0, delimiter='\t').values
    else:
        raise ValueError('Wrong Data')

    # generating drug-disease association and cross-validation, train-test data.

    num_drugs = association_matrix.shape[0]
    num_diseases = association_matrix.shape[1]
    num_nodes = num_diseases + num_drugs

    kfold = KFold(n_splits=10, shuffle=True, random_state=1024)
    pos_row, pos_col = np.nonzero(association_matrix)  # row = drugs / col = diseases
    neg_row, neg_col = np.nonzero(1 - association_matrix)  # row = drugs / col = diseases
    assert len(pos_row) + len(neg_row) == np.prod(association_matrix.shape)

    pos_row, neg_row = pos_row + num_diseases, neg_row + num_diseases

    pos_folds, neg_folds = list(kfold.split(pos_row)), list(kfold.split(neg_row))
    pos_fold, neg_fold = pos_folds[args.fold_idx], neg_folds[args.fold_idx]

    (train_pos_idx, test_pos_idx) = pos_fold
    (train_neg_idx, test_neg_idx) = neg_fold

    train_pos_edge = np.stack([pos_row[train_pos_idx], pos_col[train_pos_idx]])
    train_pos_values = [1] * len(train_pos_edge[0])
    train_neg_edge = np.stack([neg_row[train_neg_idx], neg_col[train_neg_idx]])
    train_neg_values = [0] * len(train_neg_edge[0])

    test_pos_edge = np.stack([pos_row[test_pos_idx], pos_col[test_pos_idx]])
    test_pos_values = [1] * len(test_pos_edge[0])
    test_neg_edge = np.stack([neg_row[test_neg_idx], neg_col[test_neg_idx]])
    test_neg_values = [0] * len(test_neg_edge[0])

    train_pos_edge = torch.tensor(train_pos_edge)
    train_neg_edge = torch.tensor(train_neg_edge)
    test_pos_edge = torch.tensor(test_pos_edge)
    test_neg_edge = torch.tensor(test_neg_edge)

    # We have to make csr_matrix for using GARNET.
    disease_adj = csr_matrix(disease_sim_features)
    drug_adj = csr_matrix(drug_sim_features)

    rr_disease_adj = rrSVD(disease_adj, None, r=args.disease_rank, k=args.num_k, use_feature=False)
    rr_drug_adj = rrSVD(drug_adj, None, r=args.drug_rank, k=args.num_k, use_feature=False)

    rr_disease_graph = nx.from_scipy_sparse_matrix(rr_disease_adj)
    rr_drug_graph = nx.from_scipy_sparse_matrix(rr_drug_adj)

    G = nx.Graph()
    for i in range(num_nodes):
        G.add_node(i)

    for n1, n2 in list(rr_disease_graph.edges):
        G.add_edge(n1, n2)

    for n1, n2 in list(rr_drug_graph.edges):
        G.add_edge(n1 + num_diseases, n2 + num_diseases)

    # di-dr edge
    for i in range(train_pos_edge.shape[1]):
        n1, n2 = train_pos_edge[:, i]
        n1, n2 = int(n1), int(n2)
        G.add_edge(n1, n2)

    adj = nx.adjacency_matrix(G, sorted(G.nodes()))
    A = adj.toarray()
    edge_index, _ = dense_to_sparse(torch.tensor(A))
    data_observed = Data(edge_index=edge_index)

    # set init attribute
    x = torch.ones(num_nodes, args.embedding_dim)
    x = x.float()
    data_observed.x = x

    set_random_seed(args.seed)
    train_graphs, test_graphs = [], []

    for i in range(train_pos_edge.size(1)):
        train_graphs.append(minus_edge(data_observed, 1, train_pos_edge[:, i], args))

    # negative sampling for training.
    N = train_neg_edge.size(1)
    indices = torch.randperm(N)
    train_neg_edge = train_neg_edge[:, indices]
    train_neg_edge = train_neg_edge[:, :train_pos_edge.size(1)]

    for i in range(train_neg_edge.size(1)):
        train_graphs.append(plus_edge(data_observed, 0, train_neg_edge[:, i], args))

    for i in range(test_pos_edge.size(1)):
        test_graphs.append(plus_edge(data_observed, 1, test_pos_edge[:, i], args))

    for i in range(test_neg_edge.size(1)):
        test_graphs.append(plus_edge(data_observed, 0, test_neg_edge[:, i], args))

    train_loader = DataLoader(train_graphs, batch_size=args.batch_size, shuffle=True)
    test_loader = DataLoader(test_graphs, batch_size=args.batch_size, shuffle=False)

    return train_loader, test_loader, None



